package edu.uci.ics.pattis.ics23.collections;


import java.util.Iterator;



//In this Abstract Class, many concrete methods can be written by using
//  an iterator; generally these methods are slow and should be
//  overridden in any concrete classes that extend this one.

public abstract class AbstractCollection<E> implements Iterable<E> {
	
	//Override this method if there is a more efficient algorithm
  //  for a concrete data structure
	public boolean addAll (Iterable<? extends E> es) {
 	  boolean modified = false;
	  for (E e : es)
	    if(add(e))
		    modified = true;
  	return modified;
  }
  
 
	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
	public boolean remove (E e) {
		Iterator<E> i = iterator();
		if (e == null)
		  while (i.hasNext()) {
				E item = i.next();
				if (item == null) {
				  i.remove();
				  return true;
				}
		  }
		else
			while (i.hasNext()) {
					E item = i.next();
					if (e.equals(item)) {
					  i.remove();
					  return true;
					}
			  }
		
		return false;
	}
	
	
	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
	public boolean removeAll (Iterable<? extends E> es) {
 	  boolean modified = false;
	  for (E e : es)
	    if(remove(e))
		    modified = true;
  	return modified;
  }
  
 
	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
	public boolean retainAll (Collection<? extends E> c) {
 	  boolean modified = false;
	  for (Iterator<E> i = iterator(); i.hasNext(); /* See body */)
	    if(!c.contains(i.next())) {
	    	i.remove();
		    modified = true;
	    }
  	return modified;
  }
  
 
	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public void clear () {
	  Iterator<E> i = iterator();
	  while (i.hasNext()) {
	  	i.next();
	  	i.remove();
	  }
  }

  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public boolean contains (Object o) {
  	if (o == null)
  		for (E item : this) {
  			if (item == null)
  				return true;
  		}	else
  	 	for (E item : this)
    		if (o.equals(item))
    		  return true;
  	
    	return false;
  }
 
  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public boolean containsAll (Iterable<? extends E> es) {
	  for (E e : es)
	    if(!contains(e))
		    return false;
  	return true;
  }
  
 
	
  //Generally, inherit this method without overriding
  //Ensure that size() computes its result quickly (see below)
  public boolean isEmpty()
  {return size() == 0;}


  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  //It is OK that e is never used in the for loop
  @SuppressWarnings("unused")
  public int size() {
    int count = 0;
    for (E e : this)
    	count++;
    return count;
  }

  

  //Generally, inherit this method without overriding
  //The order of the values in the collection is NOT important
  public int hashCode() {
  	int answer = 0;
   	for (E e : this)
   		answer += (e==null ? 0 : e.hashCode());
	  return answer;
  }

  
 
  //Generally, inherit this method without overriding
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public E[] toArray() {
  	E[] result = (E[]) new Object[size()];
  	int i = 0;
	  for (E e : this)
	    result[i++] = e;
	  return result;
	}

  

  //Override this method to add just the name of the concrete class
  //  and/or to illustrate the data structure used in the implementation
  //  without relying on an iterator for that class
  public String toString() {
  	StringBuffer answer = new StringBuffer();
    for (Iterator<E> i = iterator(); i.hasNext(); /*See i.next() in body*/) {
      E e = i.next();
      answer.append( (e == this ? "(this Collection)" : String.valueOf(e)) + 
      		           (i.hasNext() ? "," : "")
                    );
    }
	  return answer.toString();
	}

  
  
  //These methods must be available in all concrete classes
	public abstract boolean       add         (E e);
  public abstract Iterator<E>   iterator    ();
  public abstract Collection<E> newEmpty    ();
  public abstract Collection<E> shallowCopy ();
  public abstract boolean       equals      (Object o);
  
  
  //Used in subclasses: see their iterators
  protected transient volatile long modCount = 0;
}
