package edu.uci.ics.pattis.ics23.collections;


import java.util.Iterator;



//In this Abstract Class, many concrete methods can be written by using
//  an iterator; generally these methods are slow and should be
//  overridden in any concrete classes that extend this one, using
//  information about how the class is implemented.

public abstract class AbstractList<E> extends AbstractCollection<E> {

	
  //Generally, inherit this method without overriding
  public boolean equals(Object o) {
  	//Checks external to values in collection
  	if (o == this)
  		return true;
  	if (!(o instanceof List<?>))
  		return false;
   	List<?> l = (List<?>) o;
  	if (l.size() != size())
  		return false;

		//Checks internal to values in collection: both the same size
		try {
		  Iterator<?> ie2 = l.iterator();
			for (E e1 : this) {
				Object e2 = ie2.next();
				if (!(e1==null ? e2==null : e1.equals(e2)))
					return false;
			}
		} catch (ClassCastException e) {
			return false;
		} catch (NullPointerException e) {
			return false;
		}

  	return true;
  }

 
 //The order of the values in the collection IS important in Lists,
 //  so this method overrides one inherited from AbstractCollection
  public int hashCode() {
  	int answer = 0;
   	for (E e : this)
   		answer = 31*answer + (e==null ? 0 : e.hashCode());
	  return answer;
  }

  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public E get( int index) throws IndexOutOfBoundsException {
    int i = 0;
    for (E e : this)
      if(i == index)
        return e;
      else 
        i++;

    throw new IndexOutOfBoundsException
    ("AbstractList.get("+index+") but size()= "+i);
  }

  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public  int indexOf (E e) {
    int i = 0;
    for (E eInList : this)
     if( e==null? eInList==null : e.equals(eInList))
       return i;
     else 
       i++;
    
    return -1;
  }

  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public int lastIndexOf (E e) {
    int answer = -1;
    int i = 0;
    for (E eInList : this) {
      if( e==null? eInList==null : e.equals(eInList))
        answer = i;
      i++;
    }
    
    return answer;
  }
 
  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public E remove (int index) throws IndexOutOfBoundsException {
    if (index < 0 || index > size()-1)
      throw new IndexOutOfBoundsException("AbstractList.remove with index: index("
          +index+") illegal for size("+size()+")");

    Iterator<E> it = iterator();
    E answer = null;
    for (int i=0; i<index; i++)
      answer = it.next();

    it.remove(); 
    return answer;
  }



  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public boolean addAll (int index, Iterable<? extends E> es) throws IndexOutOfBoundsException {
    if (index < 0 || index > size())
      throw new IndexOutOfBoundsException("AbstractList.remove with index: index("
          +index+") illegal for size("+size()+")");

    int i = index;

    boolean modified = false;
    for (E e : es) {
      add(i++,e);
      modified = true;
    } 
    return modified;
  }
 
  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public  List<E> subList (int startIndex, int stopIndex) throws IndexOutOfBoundsException {
    if (startIndex > stopIndex)
      return new ArrayList<E>();

    if (startIndex < 0 || startIndex >= size() ||
        stopIndex  < 0 || stopIndex  >= size() )
      throw new IndexOutOfBoundsException("AbstractList.subList: startIndex("+startIndex+
                   ") and/or stopIndex("+stopIndex+") illegal for size("+size()+")");
    List<E> answer = (List<E>)shallowCopy();
    
    Iterator<E> it = answer.iterator();
    for (int i=0; it.hasNext(); i++) {
      it.next();
      if (i<startIndex || i>stopIndex)
        it.remove();
    }
    
    return answer;
  }


  
  //These methods must be available in all concrete classes
	public abstract boolean add    (int index, E e)                      throws IndexOutOfBoundsException ;
  public abstract E       set    (int index, E e)                      throws IndexOutOfBoundsException ;
}
