package edu.uci.ics.pattis.ics23.collections;


import java.util.Iterator;
import java.util.NoSuchElementException;



//In this Abstract Class, many concrete methods can be written by using
//an iterator; generally these methods are slow and should be
//overridden in any concrete classes that extend this one.

public  abstract class AbstractOrderedCollection<E> implements Iterable<E> {
	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
	public boolean addAll (Iterable<? extends E> es) {
 	  boolean modified = false;
	  for (E e : es)
	    if(add(e))
		    modified = true;
  	return modified;
  }
  
 
	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
	public E remove () throws NoSuchElementException {
		Iterator<E> i = iterator();
	  if (!i.hasNext())
      throw new NoSuchElementException("AbstractOrderedCollecton.remove: empty collection");
 		E answer = i.next();
		i.remove();
		return answer;
	}
	
	
	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public void clear () {
    while (!isEmpty())
      remove();
  }

  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
	public E peek () throws NoSuchElementException {
		Iterator<E> i = iterator();
	  if (!i.hasNext())
      throw new NoSuchElementException("AbstractOrderedCollecton.peek: empty collection");
		return i.next();
	}

	
	
  //Generally, inherit this method without overriding
  //Ensure that size() computes its result quickly (see below)
  public boolean isEmpty()
  {return size() == 0;}


  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  
  @SuppressWarnings("unused")
  //It is OK that e is never used in the for loop
 public int size() {
    int count = 0;
    for (E e : this)
    	count++;
    return count;
  }

  

  //The order of the values in the collection IS important in OrderedCollections
  public int hashCode() {
  	int answer = 0;
   	for (E e : this)
   		answer = 31*answer + (e==null ? 0 : e.hashCode());
	  return answer;
  }

  
  
  //Generally, inherit this method without overriding
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
   public E[] toArray() {
  	E[] result = (E[]) new Object[size()];
  	int i = 0;
	  for (E e : this)
	    result[i++] = e;
	  return result;
	}

  

  //Override this method to add just the name of the concrete class
  //  and/or to illustrate the data structure used in the implementation
  //  without relying on an iterator for that class
  public String toString() {
  	StringBuffer answer = new StringBuffer();
    for (Iterator<E> i = iterator(); i.hasNext(); /*See i.next() in body*/) {
      E e = i.next();
      answer.append( (e == this ? "(this Collection)" : String.valueOf(e)) + 
      		           (i.hasNext() ? "," : "")
                    );
    }
	  return answer.toString();
	}
 
  
  //These methods must be available in all concrete classes
  public abstract boolean              add         (E e);
  public abstract Iterator<E>          iterator    ();
  public abstract OrderedCollection<E> newEmpty    ();
  public abstract OrderedCollection<E> shallowCopy ();
  public abstract boolean              equals      (Object o);
 
  
  
  //Used in subclasses: see their iterators
  protected transient volatile long modCount = 0;
}
