package edu.uci.ics.pattis.ics23.collections;


import java.util.Iterator;
import java.util.Random;
import java.util.Comparator;
import java.util.Arrays;

import edu.uci.ics.pattis.introlib.ReverseAComparator;



public class Collections {

  private static Random randGen = new Random();

  
  public static <E> void swap(List<E> l, int i1, int i2) {
    E t = l.get(i1);
    l.set(i1,l.get(i2));
    l.set(i2,t);
  }
  
  
  
  public static <E> void reverse(List<E> l) {
    int size = l.size();
    for (int i=0; i<size/2; i++)
      swap(l, i, size-i-1);
  }
  
  
  
  public static <E> int frequency(Iterable<E> i, E value) {
    int count = 0;
    for (E e : i)
      if ( (value==null? e==null : value.equals(e)) )
        count++;

    return count;
  }
  
  
  
  public static <E> boolean uniqueElements(Iterable<E> i) {
    for (E e : i)
      if ( frequency(i,e) > 1 )
        return false;

    return true;
  }
  
  
  
  public static <E extends Comparable<? super E>> E max(Iterable<E> i) {
    E answer = null;
    for (E e : i)
      if (answer == null || answer.compareTo(e) < 0)
        answer = e;
    return answer; 
  }
 
  
  
  public static <E> E max(Iterable<E> i, Comparator<? super E>  c) {
    E answer = null;
    for (E e : i)
      if (answer == null || c.compare(answer,e) < 0)
        answer = e;
    return answer; 
  }
 
  
  
  public static <E extends Comparable<? super E>> E min(Iterable<E> i) {
    E answer = null;
    for (E e : i)
      if (answer == null || answer.compareTo(e) > 0)
        answer = e;
    return answer; 
  }
 
  
  
  public static <E> E min(Iterable<E> i, Comparator<? super E>  c) {
    E answer = null;
    for (E e : i)
      if (answer == null || c.compare(answer,e) > 0)
        answer = e;
    return answer; 
  }
 
  
  
  @SuppressWarnings("unchecked")
  public static <E extends Comparable<? super E>> void sort(List<E> l) {
    Object[] a = l.toArray();
    Arrays.sort(a);
    for (int i=0; i<a.length; i++)
      l.set(i,(E)a[i]);
  }
 
  
  
  @SuppressWarnings("unchecked")
  public static <E> void sort(List<E> l, Comparator<? super E>  c) {
    Object[] a = l.toArray();
    Arrays.sort(a,(Comparator)c);
    for (int i=0; i<a.length; i++)
      l.set(i,(E)a[i]);
  }
  
  
  
  public static <E> void shuffle(List<E> l) {
    int size = l.size();
    for (int i=size; i>1; i--)
      swap(l, i-1, randGen.nextInt(i));
  }
  
  
  
 public static <E> List<E> filter (List<E> l, Decision<E> d) {
    List<E> answer = (List<E>)l.newEmpty();
    
    for (E e : l)
      if (d.isOK(e))
        answer.add(e);
    
    return answer;
  }
  
  
  
  @SuppressWarnings("unchecked")
  public static <Domain,Range> List<Range> transform (List<Domain> l, FunctionUnary<Domain,Range> f) {
    List<Range> answer = (List<Range>)l.newEmpty();
    
    for (Domain x : l)
      answer.add(f.evaluate(x));
    
    return answer;
  }


  public static <Domain,Range> Range accumulateRight (List<Domain> l, FunctionBinary<Range,Domain,Range> f, Range identity) {
    Range answer = identity;
    
    for (int i=0; i<l.size(); i++)
      answer = f.evaluate(answer,l.get(i));
     
    return answer;
  }
  
  
  public static <Domain,Range> Range accumulateLeft (List<Domain> l, FunctionBinary<Domain,Range,Range> f, Range identity) {
    Range answer = identity;
    
    for (int i=l.size()-1; i>=0; i--)
      answer = f.evaluate(l.get(i),answer);
    
    return answer;
  }
  
  
 
  public static <E> Collection<E> constantCollection(final Collection<E> c) {
    return new Collection<E>() {
      private final Collection<E> base = c;
      //Commands
      public boolean add       (E e)                        {throw new UnsupportedOperationException("constantCollection.add");}
      public boolean addAll    (Iterable<? extends E> es)   {throw new UnsupportedOperationException("constantCollection.addAll");}
      public boolean remove    (Object o)                   {throw new UnsupportedOperationException("constantCollection.remove");}
      public boolean removeAll (Iterable<? extends E> es)   {throw new UnsupportedOperationException("constantCollection.removeAll");}
      public boolean retainAll (Collection<? extends E> c)  {throw new UnsupportedOperationException("constantCollection.retainAll");}
      public void    clear     ()                           {throw new UnsupportedOperationException("constantCollection.clear");}
      
      //Queries
      public boolean contains    (Object o)                 {return base.contains(o);}
      public boolean containsAll (Iterable<? extends E> es) {return base.containsAll(es);}
      public boolean isEmpty     ()                         {return base.isEmpty();}
      public int     size        ()                         {return base.size();}
      
      //Miscellaneous
      public Iterator<E> iterator() {
	return new Iterator<E>() {
	  Iterator<? extends E> i = base.iterator();
          public boolean hasNext() {return i.hasNext();}
	  public E       next   () {return i.next();}
  	  public void    remove () {throw new UnsupportedOperationException();}
	};
      }
      public E[]           toArray     ()                   {return base.toArray();}
      public Collection<E> newEmpty    ()                   {return base.newEmpty();}
      public Collection<E> shallowCopy ()                   {return base.shallowCopy();}
     
      //Inherited (should override)
      public String       toString ()                       {return "constantCollection["+base.toString()+"]";}
      public boolean      equals   (Object o)               {return base.equals(o);}
      public int          hashCode ()                       {return base.hashCode();}
    };
  }
 
  
  
  public static <E> Set<E> constantSet(final Set<E> s) {
    return new Set<E>() {
      private final Set<E> base = s;
      //Commands
      public boolean add       (E e)                        {throw new UnsupportedOperationException("constantSet.add");}
      public boolean addAll    (Iterable<? extends E> es)   {throw new UnsupportedOperationException("constantSet.addAll");}
      public boolean remove    (Object o)                   {throw new UnsupportedOperationException("constantSet.remove");}
      public boolean removeAll (Iterable<? extends E> es)   {throw new UnsupportedOperationException("constantCollection.removeAll");}
      public boolean retainAll (Collection<? extends E> c)  {throw new UnsupportedOperationException("constantSett.retainAll");}
      public void    clear     ()                           {throw new UnsupportedOperationException("constantSet.clear");}
      
      //Queries
      public boolean contains    (Object o)                 {return base.contains(o);}
      public boolean containsAll (Iterable<? extends E> es) {return base.containsAll(es);}
      public boolean isEmpty     ()                         {return base.isEmpty();}
      public int     size        ()                         {return base.size();}
      
      //Miscelaneous
      public Iterator<E> iterator() {
	return new Iterator<E>() {
	  Iterator<? extends E> i = base.iterator();
          public boolean hasNext() {return i.hasNext();}
	  public E       next   () {return i.next();}
  	  public void    remove () {throw new UnsupportedOperationException();}
	};
      }
      public E[]           toArray     ()                   {return base.toArray();}
      public Collection<E> newEmpty    ()                   {return base.newEmpty();}
      public Collection<E> shallowCopy ()                   {return base.shallowCopy();}
     
      //Inherited (should override)
      public String       toString ()                       {return "constantSet["+base.toString()+"]";}
      public boolean      equals   (Object o)               {return base.equals(o);}
      public int          hashCode ()                       {return base.hashCode();}
    };
  }
  
  

  public static <E> List<E> constantList(final List<E> l) {
    return new List<E>() {
      private final List<E> base = l;
      
      //Commands
      public boolean add       (E e)                                  {throw new UnsupportedOperationException("constantList.add");}
      public boolean addAll    (Iterable<? extends E> es)             {throw new UnsupportedOperationException("constantList.addAll");}
      public boolean remove    (Object o)                             {throw new UnsupportedOperationException("constantList.remove");}
      public boolean removeAll (Iterable<? extends E> es)             {throw new UnsupportedOperationException("constantList.removeAll");}
      public boolean retainAll (Collection<? extends E> c)            {throw new UnsupportedOperationException("constantList.retainAll");}
      public void    clear     ()                                     {throw new UnsupportedOperationException("constantList.clear");}
      public boolean add       (int index, E e)                       {throw new UnsupportedOperationException("constantList.add indexed");}
      public E       remove    (int index)                            {throw new UnsupportedOperationException("constantList.remove indexed");}
      public E       set       (int index, E e)                       {throw new UnsupportedOperationException("constantList.set");}
      public boolean addAll     (int index, Iterable<? extends E> es) {throw new UnsupportedOperationException("constantList.addAll indexed");}
     
      //Queries
      public boolean contains    (Object o)                           {return base.contains(o);}
      public boolean containsAll (Iterable<? extends E> es)           {return base.containsAll(es);}
      public boolean isEmpty     ()                                   {return base.isEmpty();}
      public int     size        ()                                   {return base.size();}
      public E       get         (int index)                          {return base.get(index);}
      public int     indexOf     (E e)                                {return base.indexOf(e);}
      public int     lastIndexOf (E e)                                {return base.lastIndexOf(e);}
      public List<E> subList     (int startIndex, int stopIndex)      {return base.subList(startIndex,stopIndex);}
      
      //Miscellaneous
      public Iterator<E> iterator() {
	return new Iterator<E>() {
	  Iterator<? extends E> i = base.iterator();
          public boolean hasNext() {return i.hasNext();}
	  public E       next   () {return i.next();}
  	  public void    remove () {throw new UnsupportedOperationException();}
	};
      }
      public E[]           toArray     ()                   {return base.toArray();}
      public Collection<E> newEmpty    ()                   {return base.newEmpty();}
      public Collection<E> shallowCopy ()                   {return base.shallowCopy();}
     
      //Inherited (should override)
      public String       toString ()                       {return "constantList["+base.toString()+"]";}
      public boolean      equals   (Object o)               {return base.equals(o);}
      public int          hashCode ()                       {return base.hashCode();}
    };
  }


  
  public static <K,V> Map<K,V> constantMap(final Map<K,V> m) {
    return new Map<K,V>() {
      private final Map<K,V>base = m;
      //Commands
      public V    put    (K key, V value)                      {throw new UnsupportedOperationException("constantMap.put");}
      public void putAll (Iterable<Entry<K,V>> es)             {throw new UnsupportedOperationException("constantMap.putAll");}
      public V    remove (Object key)                          {throw new UnsupportedOperationException("constantMap.remove");}
      public void clear  ()                                    {throw new UnsupportedOperationException("constantMap.clear");}

      //Queries
      public V                    get          (Object key)    {return base.get(key);}
      public boolean              containsKey  (Object key)    {return base.containsKey(key);}
      public boolean              containsValue(Object value)  {return base.containsValue(value);}
      public Iterable<Entry<K,V>> entries      (){
        return new Iterable<Entry<K,V>>() {
          public Iterator<Entry<K,V>> iterator () {
            return new Iterator<Entry<K,V>>() {
              Iterator<Entry<K,V>> i = base.entries().iterator();
              public boolean    hasNext() {return i.hasNext();}
              public Entry<K,V> next   () {return i.next();}
              public void       remove () {throw new UnsupportedOperationException();}
            };
          }
        };
      }
      public Iterable<K>          keys         () {
        return new Iterable<K>() {
          public Iterator<K> iterator () {
            return new Iterator<K>() {
              Iterator<? extends K> i = base.keys().iterator();
              public boolean hasNext() {return i.hasNext();}
              public K       next   () {return i.next();}
              public void    remove () {throw new UnsupportedOperationException();}
            };
          }
        };
      }
      public Iterable<V>          values       () {
        return new Iterable<V>() {
          public Iterator<V> iterator () {
            return new Iterator<V>() {
              Iterator<? extends V> i = base.values().iterator();
              public boolean hasNext() {return i.hasNext();}
              public V       next   () {return i.next();}
              public void    remove () {throw new UnsupportedOperationException();}
            };
          }
        };
      }
      public boolean              isEmpty      ()              {return base.isEmpty();}
      public int                  size         ()              {return base.size();}
      
      //Miscellaneous
      public Entry<K,V>[] toArray     ()                       {return base.toArray();}
      public Map<K,V>     newEmpty    ()                       {return base.newEmpty();}
      public Map<K,V>     shallowCopy ()                       {return base.shallowCopy();}
      
      //Inherited (should override)
      public String       toString ()                          {return "constantMap["+base.toString()+"]";}
      public boolean      equals   (Object o)                  {return base.equals(o);}
      public int          hashCode ()                          {return base.hashCode();}
    };
  }


  
  public static <E> OrderedCollection<E> constantOrderedCollection(final OrderedCollection<E> o) {
    return new OrderedCollection<E>() {
      private final OrderedCollection<E> base = o;
      //Commands
      public boolean add       (E e)                        {throw new UnsupportedOperationException("constantOrderedCollection.add");}
      public boolean addAll    (Iterable<? extends E> es)   {throw new UnsupportedOperationException("constantOrderedCollection.addAll");}
      public E       remove    ()                           {throw new UnsupportedOperationException("constantOrderedCollection.remove");}
      public void    clear     ()                           {throw new UnsupportedOperationException("constantOrderedCollection.clear");}
      
      //Queries
      public E       peek    ()                             {return base.peek();}
      public boolean isEmpty ()                             {return base.isEmpty();}
      public int     size    ()                             {return base.size();}
      
      //Miscelaneous
      public Iterator<E> iterator() {
	return new Iterator<E>() {
	  Iterator<? extends E> i = base.iterator();
          public boolean hasNext() {return i.hasNext();}
	  public E       next   () {return i.next();}
  	  public void    remove () {throw new UnsupportedOperationException();}
	};
      }
      public E[]                  toArray     ()            {return base.toArray();}
      public OrderedCollection<E> newEmpty    ()            {return base.newEmpty();}
      public OrderedCollection<E> shallowCopy ()            {return base.shallowCopy();}
     
      //Inherited (should override)
      public String       toString ()                       {return "constantOrderedCollection["+base.toString()+"]";}
      public boolean      equals   (Object o)               {return base.equals(o);}
      public int          hashCode ()                       {return base.hashCode();}
    };
  }
  
  
  
  public static <E> PriorityQueue<E> constantPriorityQueue(final PriorityQueue<E> p) {
    return new PriorityQueue<E>() {
      private final PriorityQueue<E> base = p;
      //Commands
      public boolean add       (E e)                      {throw new UnsupportedOperationException("constantPriorityQueue.add");}
      public boolean addAll    (Iterable<? extends E> es) {throw new UnsupportedOperationException("constantPriorityQueue.addAll");}
      public E       remove    ()                         {throw new UnsupportedOperationException("constantPriorityQueue.remove");}
      public void    clear     ()                         {throw new UnsupportedOperationException("constantPriorityQueue.clear");}
      public void    merge     (PriorityQueue<E> other)   {throw new UnsupportedOperationException("constantPriorityQueue.merge");}
      
      //Queries
      public E       peek    ()                           {return base.peek();}
      public boolean isEmpty ()                           {return base.isEmpty();}
      public int     size    ()                           {return base.size();}
      
      //Miscelaneous
      public Iterator<E> iterator() {
	return new Iterator<E>() {
	  Iterator<? extends E> i = base.iterator();
          public boolean hasNext() {return i.hasNext();}
	  public E       next   () {return i.next();}
  	  public void    remove () {throw new UnsupportedOperationException();}
	};
      }
      public E[]                  toArray     ()          {return base.toArray();}
      public OrderedCollection<E> newEmpty    ()          {return base.newEmpty();}
      public OrderedCollection<E> shallowCopy ()          {return base.shallowCopy();}
     
      //Inherited (should override)
      public String       toString ()                     {return "constantPriorityQueue["+base.toString()+"]";}
      public boolean      equals   (Object o)             {return base.equals(o);}
      public int          hashCode ()                     {return base.hashCode();}
    };
  }
  
  
  
  public static <E> Queue<E> constantQueue(final Queue<E> q) {
    return new Queue<E>() {
      private final Queue<E> base = q;
      //Commands
      public boolean add       (E e)                      {throw new UnsupportedOperationException("constantQueue.add");}
      public boolean addAll    (Iterable<? extends E> es) {throw new UnsupportedOperationException("constantQueue.addAll");}
      public E       remove    ()                         {throw new UnsupportedOperationException("constantQueue.remove");}
      public void    clear     ()                         {throw new UnsupportedOperationException("constantQueue.clear");}
      
      //Queries
      public E       peek    ()                           {return base.peek();}
      public boolean isEmpty ()                           {return base.isEmpty();}
      public int     size    ()                           {return base.size();}
      
      //Miscelaneous
      public Iterator<E> iterator() {
	return new Iterator<E>() {
	  Iterator<? extends E> i = base.iterator();
          public boolean hasNext() {return i.hasNext();}
	  public E       next   () {return i.next();}
  	  public void    remove () {throw new UnsupportedOperationException();}
	};
      }
      public E[]                  toArray     ()          {return base.toArray();}
      public OrderedCollection<E> newEmpty    ()          {return base.newEmpty();}
      public OrderedCollection<E> shallowCopy ()          {return base.shallowCopy();}
     
      //Inherited (should override)
      public String       toString ()                     {return "constantPriorityQueue["+base.toString()+"]";}
      public boolean      equals   (Object o)             {return base.equals(o);}
      public int          hashCode ()                     {return base.hashCode();}
    };
  }
  
  
  
  public static <E> Stack<E> constantStack(final Stack<E> s) {
    return new Stack<E>() {
      private final Stack<E> base = s;
      //Commands
      public boolean add       (E e)                      {throw new UnsupportedOperationException("constantStack.add");}
      public boolean addAll    (Iterable<? extends E> es) {throw new UnsupportedOperationException("constantStack.addAll");}
      public E       remove    ()                         {throw new UnsupportedOperationException("constantStack.remove");}
      public void    clear     ()                         {throw new UnsupportedOperationException("constantStack.clear");}
       
      //Queries
      public E       peek    ()                           {return base.peek();}
      public boolean isEmpty ()                           {return base.isEmpty();}
      public int     size    ()                           {return base.size();}
      
      //Miscelaneous
      public Iterator<E> iterator() {
	return new Iterator<E>() {
	  Iterator<? extends E> i = base.iterator();
          public boolean hasNext() {return i.hasNext();}
	  public E       next   () {return i.next();}
  	  public void    remove () {throw new UnsupportedOperationException();}
	};
      }
      public E[]                  toArray     ()          {return base.toArray();}
      public OrderedCollection<E> newEmpty    ()          {return base.newEmpty();}
      public OrderedCollection<E> shallowCopy ()          {return base.shallowCopy();}
     
      //Inherited (should override)
      public String       toString ()                     {return "constantStack["+base.toString()+"]";}
      public boolean      equals   (Object o)             {return base.equals(o);}
      public int          hashCode ()                     {return base.hashCode();}
    };
  }

  
  
  
  
  //Easy way to create a String of single characters
  private static String[] makeStringArray(String s) {
    String[] answer = new String[s.length()];
    for (int i=0; i<s.length(); i++)
      answer[i] = ""+s.charAt(i);
    return answer;
  }
  
  
  //For student debugging purposes: e.g., if the second JUnit test below
  //  assertEquals("remove failed" , "a", q.remove());
  //  assertArrayEquals("queue remaining after remove failed", makeStringArray("bcde")}, q.toArray());
  //fails, you can examine the array values (the values left in the queue after a remove) by
  //  System.out.println( arrayToString(q.toArray()) );
  private static String arrayToString (Object[] q) {
    StringBuffer answer = new StringBuffer();
    for (int i=0; i<q.length; i++)
      answer.append("["+i+"]:" + q[i] + (i==q.length-1?"":","));
    return answer.toString();
  }
  
  public static void main (String[] s) {
     List<String> l = new ArrayList<String>(makeStringArray("abcdefghijklmnopqrstuvwxyz"));
    for (int i=0; i< 10; i++) {
      shuffle(l);
      System.out.println(arrayToString(l.toArray()));
      System.out.println(min(l) + ":"+ max(l));
      sort(l, new ReverseAComparator<String>(
          new Comparator<String>()
          {public int compare(String s1, String s2){return s1.compareTo(s2);}}));
      System.out.println(arrayToString(l.toArray()));
      reverse(l);
      System.out.println(arrayToString(l.toArray()));
   }
   List<String> a = new ArrayList<String>(makeStringArray("abecde"));
   System.out.println(a.size());
//   a = constantList(a);
   System.out.println(a);
   System.out.println("unique:" + uniqueElements(a));
   System.out.println(a.add("xyz"));
   System.out.println(
       transform(a, new FunctionUnary<String,Integer>()
       {public Integer evaluate(String x) {return x.length();}}));
  }
     
  

}
